"""Radarr Common Models."""
# pylint: disable=invalid-name, too-many-instance-attributes
from __future__ import annotations

from dataclasses import dataclass

from .base import APIResponseType, BaseModel
from .common import _CommonAttrs


@dataclass(init=False)
class _RadarrMovieQualityRevision(BaseModel):
    """Movie file revision attributes."""

    isRepack: bool | None = None
    real: int | None = None
    version: int | None = None


@dataclass(init=False)
class _RadarrCommon4(BaseModel):
    """Movie file language attributes."""

    id: int | None = None
    name: str | None = None


@dataclass(init=False)
class _RadarrMovieQualityInfo(_RadarrCommon4):
    """Movie file quality attributes."""

    modifier: str | None = None
    resolution: int | None = None
    source: str | None = None


@dataclass(init=False)
class _RadarrMovieQuality(BaseModel):
    """Movie file quality attributes."""

    quality: _RadarrMovieQualityInfo | None = None
    revision: _RadarrMovieQualityRevision | None = None

    def __post_init__(self):
        super().__post_init__()
        self.quality = _RadarrMovieQualityInfo(self.quality) or {}
        self.revision = _RadarrMovieQualityRevision(self.revision) or {}


@dataclass(init=False)
class _RadarrMovieFileMediaInfo(_CommonAttrs, BaseModel):
    """Movie file media attributes."""

    audioAdditionalFeatures: str | None = None


@dataclass(init=False)
class _RadarrMovieImages(BaseModel):
    """Movie images attributes."""

    coverType: str | None = None
    remoteUrl: str | None = None
    url: str | None = None


@dataclass(init=False)
class _RadarrMovieRatings(BaseModel):
    """Movie ratings attributes."""

    value: float | None = None
    votes: int | None = None


@dataclass(init=False)
class _RadarrMovieCollection(BaseModel):
    """Movie collection attributes."""

    images: list[_RadarrMovieImages] | None = None
    name: str | None = None
    tmdbId: int | None = None

    def __post_init__(self):
        super().__post_init__()
        self.images = [_RadarrMovieImages(image) for image in self.images or []]


@dataclass(init=False)
class _RadarrMovieFields(BaseModel):
    """Movie fields attributes."""

    advanced: bool | None = None
    helpText: str | None = None
    label: str | None = None
    name: str | None = None
    order: int | None = None
    type: str | None = None
    value: str | None = None


@dataclass(init=False)
class _RadarrCommon(BaseModel):
    """Radarr indexers attributes."""

    fields: list[_RadarrMovieFields] | None = None
    implementation: str | None = None
    implementationName: str | None = None
    infoLink: str | None = None
    name: str | None = None

    def __post_init__(self):
        super().__post_init__()
        self.fields = [_RadarrMovieFields(field) for field in self.fields or []]


@dataclass(init=False)
class _RadarrMovieSpecifications(_RadarrCommon):
    """Movie specifications attributes."""

    negate: bool | None = None
    required: bool | None = None


@dataclass(init=False)
class _RadarrMovieCustomFormats(_RadarrCommon4):
    """Movie custom formats attributes."""

    includeCustomFormatWhenRenaming: bool | None = None
    specifications: list[_RadarrMovieSpecifications] | None = None

    def __post_init__(self):
        super().__post_init__()
        self.specifications = [
            _RadarrMovieSpecifications(spec) for spec in self.specifications or []
        ]


@dataclass(init=False)
class _RadarrCommon2(BaseModel):
    """Radarr common attributes."""

    id: int | None = None
    minimumAvailability: str | None = None
    qualityProfileId: int | None = None


@dataclass(init=False)
class _RadarrMovieCommon(BaseModel):
    """Movie common attributes."""

    id: int | None = None
    languages: list[_RadarrCommon4] | None = None
    quality: _RadarrMovieQuality | None = None

    def __post_init__(self):
        super().__post_init__()
        self.languages = [_RadarrCommon4(language) for language in self.languages or []]
        self.quality = _RadarrMovieQuality(self.quality) or {}


@dataclass(init=False)
class _RadarrMovieHistoryBlocklistBase(_RadarrMovieCommon):
    """Movie history/blocklist attributes."""

    _responsetype = APIResponseType.LIST

    customFormats: list[_RadarrMovieCustomFormats] | None = None
    date: str | None = None
    movieId: int | None = None
    sourceTitle: str | None = None

    def __post_init__(self):
        super().__post_init__()
        self.customFormats = [
            _RadarrMovieCustomFormats(custForm) for custForm in self.customFormats or []
        ]


@dataclass(init=False)
class _RadarrQueueStatusMessages(BaseModel):
    """Radarr queue status messages."""

    title: str | None = None
    messages: list[dict] | None = None


@dataclass(init=False)
class _RadarrNotificationMessage(BaseModel):
    """Radarr notification message attributes."""

    message: str | None = None
    type: str | None = None


@dataclass(init=False)
class _RadarrUpdateChanges(BaseModel):
    """Radarr recent updates changes attributes."""

    fixed: list[str] | None = None
    new: list[str] | None = None


@dataclass(init=False)
class _RadarrQualityProfileItems(_RadarrCommon4):
    """Radarr quality profile items attributes."""

    allowed: bool | None = None
    items: list[_RadarrQualityProfileItems] | None = None
    quality: _RadarrMovieQualityInfo | None = None

    def __post_init__(self):
        super().__post_init__()
        self.items = [_RadarrQualityProfileItems(item) for item in self.items or []]
        if isinstance(self.quality, dict):
            self.quality = _RadarrMovieQualityInfo(self.quality) or {}


@dataclass(init=False)
class _RadarrMovieFileCommon(_RadarrMovieCommon):
    """Movie file attributes."""

    dateAdded: str | None = None
    indexerFlags: int | None = None
    mediaInfo: _RadarrMovieFileMediaInfo | None = None
    movieId: int | None = None
    path: str | None = None
    qualityCutoffNotMet: bool | None = None
    relativePath: str | None = None
    releaseGroup: str | None = None
    size: int | None = None

    def __post_init__(self):
        super().__post_init__()
        self.mediaInfo = _RadarrMovieFileMediaInfo(self.mediaInfo) or {}


@dataclass(init=False)
class _RadarrCalendarMovieFile(_RadarrMovieFileCommon):
    """Calendar movie file attributes."""

    edition: str | None = None
    originalFilePath: str | None = None
    sceneName: str | None = None


@dataclass(init=False)
class _RadarrCustomFilterAttr(BaseModel):
    """Radarr custom filter attributes."""

    key: str | None = None
    type: str | None = None
    value: list[str] | None = None


@dataclass(init=False)
class _RadarrCommandBody(BaseModel):
    """Radarr command body attributes."""

    completionMessage: str | None = None
    isExclusive: bool | None = None
    isNewMovie: bool | None = None
    isTypeExclusive: bool | None = None
    lastExecutionTime: str | None = None
    lastStartTime: str | None = None
    name: str | None = None
    requiresDiskAccess: bool | None = None
    sendUpdatesToClient: bool | None = None
    suppressMessages: bool | None = None
    trigger: str | None = None
    updateScheduledTask: bool | None = None


@dataclass(init=False)
class _RadarrUnmappedRootFolder(BaseModel):
    """Radarr root unmapped folder attributes."""

    name: str | None = None
    path: str | None = None


@dataclass(init=False)
class _RadarrMovieHistoryData(BaseModel):
    """Movie history data attributes."""

    reason: str | None = None



"""Shared API."""
from __future__ import annotations

import asyncio
from copy import copy

import aiohttp
from aiohttp.client import ClientSession

from aiopyarr.models.base import BaseModel
from aiopyarr.models.sonarr import Logs

from .const import ATTR_DATA, LOGGER, HTTPMethod
from .models.host_configuration import PyArrHostConfiguration
from .models.response import PyArrResponse

from .exceptions import (  # isort:skip
    ArrAuthenticationException,
    ArrConnectionException,
    ArrException,
    ArrResourceNotFound,
)


class RequestClient:
    """Base class for API Client."""

    _close_session = False
    _from_aenter = False

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host_configuration: PyArrHostConfiguration | None = None,
        session: ClientSession | None = None,
        hostname: str | None = None,
        ipaddress: str | None = None,
        url: str | None = None,
        api_token: str | None = None,
        port: int | None = None,
        ssl: bool | None = None,
        verify_ssl: bool | None = None,
        base_api_path: str | None = None,
        request_timeout: float = 30,
        raw_response: bool = False,
        redact: bool = True,
    ) -> None:
        """Initialize."""
        if host_configuration is None:
            host_configuration = PyArrHostConfiguration(
                hostname=hostname, ipaddress=ipaddress, url=url, api_token=api_token
            )
        else:
            host_configuration = copy(host_configuration)

        if port is not None:
            host_configuration.port = port
        if ssl is not None:
            host_configuration.ssl = ssl
        if verify_ssl is not None:
            host_configuration.verify_ssl = verify_ssl
        if base_api_path is not None:
            host_configuration.base_api_path = base_api_path

        if session is None:
            session = ClientSession()
            self._close_session = True

        self._host = host_configuration
        self._session = session
        self._request_timeout = request_timeout
        self._raw_response = raw_response
        self._redact = redact

    async def __aenter__(self) -> RequestClient:
        """Async enter."""
        self._from_aenter = True
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        if self._session and self._close_session:
            await self._session.close()

    def redact_string(self, string: str) -> str:
        """Redact a api token from a string if needed."""
        if not self._redact:
            return string

        return string.replace(self._host.api_token, "[REDACTED_API_TOKEN]")

    async def _async_request(
        self,
        *args,
        params: dict | None = None,
        data: dict | None = None,
        datatype: BaseModel | None = None,
        method: HTTPMethod = HTTPMethod.GET,
    ):
        """Send API request."""
        command = args[0] if isinstance(args[0], str) else args[1]
        url = self._host.api_url(command)
        try:
            request = await self._session.request(
                method=method.value,
                url=url,
                params=params,
                json=data,
                verify_ssl=self._host.verify_ssl,
                timeout=aiohttp.ClientTimeout(self._request_timeout),
            )

            _result: dict = await request.json()

            response = PyArrResponse(
                data={ATTR_DATA: _result},
                datatype=datatype,
            )

            if request.status != 200:

                if request.status == 401:
                    raise ArrAuthenticationException(self, response.message)
                if request.status == 404:
                    raise ArrResourceNotFound(self, response.message)
                raise ArrConnectionException(
                    self,
                    f"Request for '{url}' failed with status code '{request.status}'",
                )

            LOGGER.debug("Requesting %s returned %s", self.redact_string(url), _result)

            if self._raw_response:
                return _result

        except aiohttp.ClientError as exception:
            raise ArrConnectionException(
                self,
                f"Request exception for '{url}' with - {exception}",
            ) from exception

        except asyncio.TimeoutError as ex:
            raise ArrConnectionException(self, f"Request timeout for '{url}'") from ex

        except ArrAuthenticationException as ex:
            raise ArrAuthenticationException(self, ex) from ex

        except ArrConnectionException as ex:
            raise ArrConnectionException(self, ex) from ex

        except ArrException as ex:
            raise ArrException(self, ex) from ex

        except (Exception, BaseException) as ex:
            raise ArrException(self, ex) from ex

        else:
            return response.data

    async def async_get_logs(  # pylint: disable=too-many-arguments
        self,
        page: int = 1,
        page_size: int = 10,
        sort_key: str = "time",
        sort_asc: bool = False,
        filter_key: str | None = None,
        filter_value: str = "All",
    ) -> Logs:
        """Get logs.

        Args:
            page: Specifiy page to return.
            page_size: Number of items per page.
            sort_key: Field to sort by.
            sort_asc: Sort items in ascending order.
            filter_key: Key to filter by.
            filter_value: Value of the filter.
        """
        params = {
            "page": page,
            "pageSize": page_size,
            "sortKey": sort_key,
            "sortDir": "asc" if sort_asc is True else "desc",
            "filterKey": str(filter_key),
            "filterValue": filter_value,
        }
        return await self._async_request("log", params=params, datatype=Logs)



"""Exceptions for Arr Api Client."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .request_client import RequestClient


class ArrException(Exception):
    """Base arr exception."""

    def __init__(
        self,
        client: RequestClient | None = None,
        message: str | Exception = "",
    ) -> None:
        """Initialize."""
        super().__init__(
            client.redact_string(str(message)) if client is not None else message
        )


class ArrConnectionException(ArrException):
    """Arr connection exception."""


class ArrAuthenticationException(ArrException):
    """Arr authentication exception."""


class ArrResourceNotFound(ArrException):
    """Arr resoruce not found exception."""


class ArrInvalidCommand(ArrException):
    """Arr resoruce not found exception."""



"""Sonarr Models."""
# pylint: disable=invalid-name, too-many-instance-attributes
from __future__ import annotations

from dataclasses import dataclass

from .base import APIResponseType, BaseModel

from .sonarr_common import (  # isort:skip
    _FileEpisodeQuality,
    _SonarrCommon,
    _SonarrCommon3,
    _SonarrCommon4,
    _SonarrCommon5,
    _SonarrCommon6,
    _SonarrCommon7,
    _SonarrCutoff,
    _SonarrEpisodeFile,
    _SonarrHistoryRecord,
    _SonarrHistoryRecordEpisode,
    _SonarrHistoryRecordSeries,
    _SonarrLogRecord,
    _SonarrParseEpisodeInfo,
    _SonarrQualityProfileValueItems,
    _SonarrQualitySub,
    _SonarrSeasonStatistics,
    _SonarrSeries2,
    _SonarrSeriesAlternateTitle,
    _SonarrSeriesCommon,
    _SonarrSeriesCommon3,
    _SonarrWantedMissingRecord,
    _SonarrWantedMissingSeriesSeason,
)


@dataclass(init=False)
class SonarrCalendar(_SonarrCommon5):
    """Sonarr calendar attributes."""

    _responsetype = APIResponseType.LIST

    downloading: bool | None = None
    series: SonarrSeries | None = None
    sceneEpisodeNumber: int | None = None
    sceneSeasonNumber: int | None = None
    tvDbEpisodeId: int | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.series = SonarrSeries(self.series) or {}


@dataclass(init=False)
class SonarrCommand(_SonarrCommon4):
    """Sonarr command attributes."""

    sendUpdatesToClient: bool | None = None
    startedOn: str | None = None
    state: str | None = None
    stateChangeTime: str | None = None


@dataclass(init=False)
class SonarrEpisode(_SonarrCommon5):
    """Sonarr episode attributes."""

    absoluteEpisodeNumber: int | None = None
    episodeFile: _SonarrEpisodeFile | None = None
    sceneEpisodeNumber: int | None = None
    sceneSeasonNumber: int | None = None
    series: _SonarrHistoryRecordSeries | None = None
    tvDbEpisodeId: int | None = None
    unverifiedSceneNumbering: bool | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.episodeFile = _SonarrEpisodeFile(self.episodeFile) or {}
        self.series = _SonarrHistoryRecordSeries(self.series) or {}


@dataclass(init=False)
class SonarrEpisodeFile(_SonarrEpisodeFile):
    """Sonarr episode file attributes."""


@dataclass(init=False)
class SonarrEpisodeFileQuailty(_SonarrCommon6):
    """Required input for updating episode file quality."""

    quality: _FileEpisodeQuality | None = None


@dataclass(init=False)
class SonarrHistory(_SonarrCommon):
    """Sonarr history attributes."""

    records: list[_SonarrHistoryRecord] | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.records = [_SonarrHistoryRecord(record) for record in self.records or []]


@dataclass(init=False)
class SonarrWantedMissing(_SonarrCommon):
    """Sonarr wanted missing attributes."""

    records: list[_SonarrWantedMissingRecord] | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.records = [
            _SonarrWantedMissingRecord(record) for record in self.records or []
        ]


@dataclass(init=False)
class SonarrQueueNew(_SonarrCommon6):
    """Sonarr queue attributes."""

    _responsetype = APIResponseType.LIST

    downloadId: str | None = None
    episode: _SonarrHistoryRecordEpisode | None = None
    estimatedCompletionTime: str | None = None
    protocol: str | None = None
    quality: _SonarrQualitySub | None = None
    series: _SonarrHistoryRecordSeries | None = None
    size: int | None = None
    sizeleft: int | None = None
    status: str | None = None
    statusMessages: list[str] | None = None
    timeleft: str | None = None
    title: str | None = None
    trackedDownloadStatus: str | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.episode = _SonarrHistoryRecordEpisode(self.episode) or {}
        self.quality = _SonarrQualitySub(self.quality) or {}
        if isinstance(self.series, dict):
            self.series = _SonarrHistoryRecordSeries(self.series) or {}


@dataclass(init=False)
class SonarrParse(BaseModel):
    """Sonarr parse attributes."""

    episodes: list[SonarrEpisode] | None = None
    parsedEpisodeInfo: _SonarrParseEpisodeInfo | None = None
    series: _SonarrSeries2 | None = None
    title: str | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.parsedEpisodeInfo = _SonarrParseEpisodeInfo(self.parsedEpisodeInfo) or {}


@dataclass(init=False)
class SonarrQualityProfile(_SonarrCommon4):
    """Sonarr quality profile attributes."""

    _responsetype = APIResponseType.LIST

    cutoff: _SonarrCutoff | None = None
    items: list[_SonarrQualityProfileValueItems] | None = None


@dataclass(init=False)
class SonarrRelease(_SonarrCommon3, _SonarrCommon7):
    """Sonarr release attributes."""

    publishDate: str | None = None


@dataclass(init=False)
class SonarrRootFolder(_SonarrCommon6):
    """Sonarr root folder attributes."""

    _responsetype = APIResponseType.LIST

    freeSpace: int | None = None
    path: str | None = None
    unmappedFolders: list[str] | None = None


@dataclass(init=False)
class SonarrSeries(_SonarrSeasonStatistics, _SonarrHistoryRecordSeries):
    """Sonarr series attributes."""

    alternateTitles: list[_SonarrSeriesAlternateTitle] | None = None
    languageProfileId: int | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.alternateTitles = [
            _SonarrSeriesAlternateTitle(altTit) for altTit in self.alternateTitles or []
        ]


@dataclass(init=False)
class SonarrSeriesLookup(_SonarrSeriesCommon):
    """Sonarr series lookup attributes."""

    _responsetype = APIResponseType.LIST

    remotePoster: str | None = None


@dataclass(init=False)
class SonarrSystemStatus(BaseModel):
    """Sonarr system status attributes."""

    appData: str | None = None
    authentication: bool | None = None
    branch: str | None = None
    buildTime: str | None = None
    isAdmin: bool | None = None
    isDebug: bool | None = None
    isLinux: bool | None = None
    isMono: bool | None = None
    isProduction: bool | None = None
    isUserInteractive: bool | None = None
    isWindows: bool | None = None
    osVersion: str | None = None
    startOfWeek: int | None = None
    startupPath: str | None = None
    urlBase: str | None = None
    version: str | None = None


@dataclass(init=False)
class SonarrSystemBackup(_SonarrCommon4):
    """Sonarr system backup attributes."""

    _responsetype = APIResponseType.LIST

    path: str | None = None
    time: str | None = None
    type: str | None = None


@dataclass(init=False)
class SonarrTag(_SonarrCommon6):
    """Sonarr tag attributes."""

    label: str | None = None


@dataclass(init=False)
class SonarrSeriesUpdateParams(_SonarrSeriesCommon3):
    """Sonarr series update parameters."""

    profileId: int | None = None
    seasons: list[_SonarrWantedMissingSeriesSeason] | None = None


@dataclass(init=False)
class Logs(_SonarrCommon):
    """Log attributes."""

    records: list[_SonarrLogRecord] | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.records = [_SonarrLogRecord(record) for record in self.records or []]



"""Radarr Models."""
# pylint: disable=invalid-name, too-many-instance-attributes
from __future__ import annotations

from dataclasses import dataclass

from .base import APIResponseType, BaseModel

from .radarr_common import (  # isort:skip
    _RadarrCalendarMovieFile,
    _RadarrCommandBody,
    _RadarrCommon,
    _RadarrCommon2,
    _RadarrCommon4,
    _RadarrCustomFilterAttr,
    _RadarrMovieCollection,
    _RadarrMovieCustomFormats,
    _RadarrMovieFileCommon,
    _RadarrMovieHistoryBlocklistBase,
    _RadarrMovieHistoryData,
    _RadarrMovieImages,
    _RadarrMovieQuality,
    _RadarrMovieRatings,
    _RadarrNotificationMessage,
    _RadarrQualityProfileItems,
    _RadarrQueueStatusMessages,
    _RadarrUnmappedRootFolder,
    _RadarrUpdateChanges,
)


@dataclass(init=False)
class RadarrMovie(_RadarrCommon2):
    """Movie attributes."""

    added: str | None = None
    certification: str | None = None
    cleanTitle: str | None = None
    collection: _RadarrMovieCollection | None = None
    folderName: str | None = None
    genres: list[str] | None = None
    hasFile: bool | None = None
    imdbId: str | None = None
    images: list[_RadarrMovieImages] | None = None
    inCinemas: str | None = None
    isAvailable: bool | None = None
    monitored: bool | None = None
    overview: str | None = None
    path: str | None = None
    physicalRelease: str | None = None
    ratings: _RadarrMovieRatings | None = None
    rootFolderPath: str | None = None
    runtime: int | None = None
    sizeOnDisk: int | None = None
    sortTitle: str | None = None
    status: str | None = None
    studio: str | None = None
    tags: list[int] | None = None
    title: str | None = None
    titleSlug: str | None = None
    tmdbId: int | None = None
    website: str | None = None
    year: int | None = None
    youTubeTrailerId: str | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.images = [_RadarrMovieImages(image) for image in self.images or []]
        self.ratings = _RadarrMovieRatings(self.ratings) or {}
        if isinstance(self.collection, dict):
            self.collection = _RadarrMovieCollection(self.collection) or {}


@dataclass(init=False)
class RadarrMovieFile(_RadarrMovieFileCommon):
    """Movie file attributes."""


@dataclass(init=False)
class RadarrMovieHistory(_RadarrMovieHistoryBlocklistBase):
    """Movie file quality attributes."""

    data: _RadarrMovieHistoryData | None = None
    downloadId: str | None = None
    eventType: str | None = None
    qualityCutoffNotMet: bool | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.data = _RadarrMovieHistoryData(self.data) or {}


@dataclass(init=False)
class RadarrBlocklistMovie(_RadarrMovieHistoryBlocklistBase):
    """Blocklist movie attributes."""

    indexer: str | None = None
    protocol: str | None = None


@dataclass(init=False)
class RadarrBlocklist(BaseModel):
    """Blocklist attributes."""

    page: int | None = None
    pageSize: int | None = None
    records: list[RadarrBlocklistMovie] | None = None
    sortDirection: str | None = None
    sortKey: str | None = None
    totalRecords: int | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.records = [RadarrBlocklistMovie(record) for record in self.records or []]


@dataclass(init=False)
class RadarrQueueDetail(BaseModel):
    """Radarr queue details attributes."""

    _responsetype = APIResponseType.LIST

    customFormats: list[_RadarrMovieCustomFormats] | None = None
    downloadClient: str | None = None
    downloadId: str | None = None
    errorMessage: str | None = None
    estimatedCompletionTime: str | None = None
    id: int | None = None
    indexer: str | None = None
    languages: list[_RadarrCommon4] | None = None
    outputPath: str | None = None
    protocol: str | None = None
    quality: _RadarrMovieQuality | None = None
    size: int | None = None
    sizeleft: int | None = None
    status: str | None = None
    statusMessages: list[_RadarrQueueStatusMessages] | None = None
    timeleft: str | None = None
    title: str | None = None
    trackedDownloadStatus: str | None = None
    trackedDownloadState: str | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.customFormats = [
            _RadarrMovieCustomFormats(custForm) for custForm in self.customFormats or []
        ]
        self.languages = [_RadarrCommon4(language) for language in self.languages or []]
        self.statusMessages = [
            _RadarrQueueStatusMessages(statMsg) for statMsg in self.statusMessages or []
        ]
        self.quality = _RadarrMovieQuality(self.quality) or {}


@dataclass(init=False)
class RadarrQueue(BaseModel):
    """Radarr queue attributes."""

    page: int | None = None
    pageSize: int | None = None
    sortKey: str | None = None
    sortDirection: str | None = None
    totalRecords: int | None = None
    records: list[RadarrQueueDetail] | None = None


@dataclass(init=False)
class RadarrQueueStatus(BaseModel):
    """Radarr queue status attributes."""

    totalCount: int | None = None
    count: int | None = None
    unknownCount: int | None = None
    errors: bool | None = None
    warnings: bool | None = None
    unknownErrors: bool | None = None
    unknownWarnings: bool | None = None


@dataclass(init=False)
class RadarrIndexer(_RadarrCommon):
    """Radarr indexers attributes."""

    configContract: str | None = None
    enableAutomaticSearch: bool | None = None
    enableInteractiveSearch: bool | None = None
    enableRss: bool | None = None
    id: int | None = None
    protocol: str | None = None
    priority: int | None = None
    supportsRss: bool | None = None
    supportsSearch: bool | None = None
    tags: list[int] | None = None


@dataclass(init=False)
class RadarrDownloadClient(_RadarrCommon):
    """Radarr indexers attributes."""

    configContract: str | None = None
    enable: bool | None = None
    id: int | None = None
    priority: int | None = None
    protocol: str | None = None
    tags: list[int] | None = None


@dataclass(init=False)
class RadarrImportList(_RadarrCommon, _RadarrCommon2):
    """Radarr import attributes."""

    configContract: str | None = None
    enableAuto: bool | None = None
    enabled: bool | None = None
    listOrder: int | None = None
    listType: str | None = None
    rootFolderPath: str | None = None
    searchOnAdd: bool | None = None
    shouldMonitor: bool | None = None
    tags: list[int] | None = None


@dataclass(init=False)
class RadarrNotification(_RadarrCommon):
    """Radarr notification attributes."""

    configContract: str | None = None
    id: int | None = None
    includeHealthWarnings: bool | None = None
    message: _RadarrNotificationMessage | None = None
    onDelete: bool | None = None
    onDownload: bool | None = None
    onGrab: bool | None = None
    onHealthIssue: bool | None = None
    onRename: bool | None = None
    onUpgrade: bool | None = None
    supportsOnDelete: bool | None = None
    supportsOnDownload: bool | None = None
    supportsOnGrab: bool | None = None
    supportsOnHealthIssue: bool | None = None
    supportsOnRename: bool | None = None
    supportsOnUpgrade: bool | None = None
    tags: list[int] | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.message = _RadarrNotificationMessage(self.message) or {}


@dataclass(init=False)
class RadarrTag(BaseModel):
    """Radarr tag attributes."""

    id: int | None = None
    label: str | None = None


@dataclass(init=False)
class RadarrTagDetails(RadarrTag):
    """Radarr tag details attributes."""

    delayProfileIds: list[int] | None = None
    notificationIds: list[int] | None = None
    restrictionIds: list[int] | None = None
    netImportIds: list[int] | None = None
    movieIds: list[int] | None = None


@dataclass(init=False)
class Diskspace(BaseModel):
    """Radarr diskspace attributes."""

    _responsetype = APIResponseType.LIST

    freeSpace: int | None = None
    label: str | None = None
    path: str | None = None
    totalSpace: int | None = None


@dataclass(init=False)
class RadarrUIConfig(BaseModel):
    """Radarr UI config attributes."""

    calendarWeekColumnHeader: str | None = None
    enableColorImpairedMode: bool | None = None
    firstDayOfWeek: int | None = None
    id: int | None = None
    longDateFormat: str | None = None
    movieInfoLanguage: int | None = None
    movieRuntimeFormat: str | None = None
    shortDateFormat: str | None = None
    showRelativeDates: bool | None = None
    timeFormat: str | None = None


@dataclass(init=False)
class RadarrHostConfig(BaseModel):
    """Radarr host config attributes."""

    analyticsEnabled: bool | None = None
    apiKey: str | None = None
    authenticationMethod: str | None = None
    backupFolder: str | None = None
    backupInterval: int | None = None
    backupRetention: int | None = None
    bindAddress: str | None = None
    branch: str | None = None
    certificateValidation: str | None = None
    consoleLogLevel: str | None = None
    enableSsl: bool | None = None
    id: int | None = None
    launchBrowser: bool | None = None
    logLevel: str | None = None
    password: str | None = None
    port: int | None = None
    proxyBypassFilter: str | None = None
    proxyBypassLocalAddresses: bool | None = None
    proxyEnabled: bool | None = None
    proxyHostname: str | None = None
    proxyPassword: str | None = None
    proxyPort: int | None = None
    proxyType: str | None = None
    proxyUsername: str | None = None
    sslCertPassword: str | None = None
    sslCertPath: str | None = None
    sslPort: int | None = None
    urlBase: str | None = None
    updateAutomatically: bool | None = None
    updateMechanism: str | None = None
    updateScriptPath: str | None = None
    username: str | None = None


@dataclass(init=False)
class RadarrNamingConfig(BaseModel):
    """Radarr host naming config attributes."""

    colonReplacementFormat: str | None = None
    id: int | None = None
    includeQuality: bool | None = None
    movieFolderFormat: str | None = None
    renameMovies: bool | None = None
    replaceIllegalCharacters: bool | None = None
    replaceSpaces: bool | None = None
    standardMovieFormat: str | None = None


@dataclass(init=False)
class RadarrMetadataConfig(_RadarrCommon):
    """Radarr metadata consumer config attributes."""

    _responsetype = APIResponseType.LIST

    configContract: str | None = None
    enable: bool | None = None
    id: int | None = None
    tags: list[int] | None = None


@dataclass(init=False)
class RadarrSystemStatus(BaseModel):
    """Radarr system status attributes."""

    appData: str | None = None
    authentication: str | None = None
    branch: str | None = None
    buildTime: str | None = None
    isAdmin: bool | None = None
    isDebug: bool | None = None
    isDocker: bool | None = None
    isLinux: bool | None = None
    isMono: bool | None = None
    isNetCore: bool | None = None
    isOsx: bool | None = None
    isProduction: bool | None = None
    isUserInteractive: bool | None = None
    isWindows: bool | None = None
    migrationVersion: int | None = None
    mode: str | None = None
    osName: str | None = None
    osVersion: str | None = None
    packageUpdateMechanism: str | None = None
    runtimeVersion: str | None = None
    runtimeName: str | None = None
    sqliteVersion: str | None = None
    startTime: str | None = None
    startupPath: str | None = None
    urlBase: str | None = None
    version: str | None = None


@dataclass(init=False)
class RadarrHealth(BaseModel):
    """Radarr failed health check attributes."""

    _responsetype = APIResponseType.LIST

    message: str | None = None
    source: str | None = None
    type: str | None = None
    wikiUrl: str | None = None


@dataclass(init=False)
class RadarrUpdate(BaseModel):
    """Radarr recent updates attributes."""

    branch: str | None = None
    changes: _RadarrUpdateChanges | None = None
    fileName: str | None = None
    hash: str | None = None
    installable: bool | None = None
    installed: bool | None = None
    latest: bool | None = None
    releaseDate: str | None = None
    url: str | None = None
    version: str | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.changes = _RadarrUpdateChanges(self.changes) or {}


@dataclass(init=False)
class RadarrQualityProfile(_RadarrCommon4):
    """Radarr quality profile attributes."""

    _responsetype = APIResponseType.LIST

    cutoff: int | None = None
    cutoffFormatScore: int | None = None
    formatItems: list | None = None
    items: list[_RadarrQualityProfileItems] | None = None
    language: _RadarrCommon4 | None = None
    minFormatScore: int | None = None
    upgradeAllowed: bool | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.items = [_RadarrQualityProfileItems(item) for item in self.items or []]
        self.language = _RadarrCommon4(self.language) or {}


@dataclass(init=False)
class RadarrCalendar(RadarrMovie, _RadarrCommon2):
    """Radarr calendar attributes."""

    _responsetype = APIResponseType.LIST

    alternateTitles: list[str] | None = None
    digitalRelease: str | None = None
    minimumAvailability: str | None = None
    movieFile: _RadarrCalendarMovieFile | None = None
    originalTitle: str | None = None
    qualityProfileId: int | None = None
    secondaryYearSourceId: int | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        if isinstance(self.images, list):
            self.movieFile = _RadarrCalendarMovieFile(self.movieFile) or {}


@dataclass(init=False)
class RadarrCustomFilter(BaseModel):
    """Radarr custom filter attributes."""

    _responsetype = APIResponseType.LIST

    id: int | None = None
    filters: list[_RadarrCustomFilterAttr] | None = None
    label: str | None = None
    type: str | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.filters = [
            _RadarrCustomFilterAttr(filter) for filter in self.filters or []
        ]


@dataclass(init=False)
class RadarrRemotePathMapping(BaseModel):
    """Radarr remote path mapping attributes."""

    _responsetype = APIResponseType.LIST

    id: int | None = None
    host: str | None = None
    localPath: str | None = None
    remotePath: str | None = None


@dataclass(init=False)
class RadarrRootFolder(BaseModel):
    """Radarr root folder attributes."""

    _responsetype = APIResponseType.LIST

    id: int | None = None
    freeSpace: int | None = None
    path: str | None = None
    unmappedFolders: list[_RadarrUnmappedRootFolder] | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.unmappedFolders = [
            _RadarrUnmappedRootFolder(unmap) for unmap in self.unmappedFolders or []
        ]


@dataclass(init=False)
class RadarrMovieEditor(BaseModel):
    """Radarr root folder attributes."""

    applyTags: str | None = None
    minimumAvailability: str | None = None
    monitored: bool | None = None
    moveFiles: bool | None = None
    movieIds: list[int] | None = None
    qualityProfileId: int | None = None
    rootFolderPath: str | None = None
    tags: list[int] | None = None


@dataclass(init=False)
class RadarrCommand(_RadarrCommon4):
    """Radarr command attributes."""

    _responsetype = APIResponseType.LIST

    body: _RadarrCommandBody | None = None
    commandName: str | None = None
    duration: str | None = None
    ended: str | None = None
    lastExecutionTime: str | None = None
    message: str | None = None
    priority: str | None = None
    queued: str | None = None
    sendUpdatesToClient: bool | None = None
    started: str | None = None
    stateChangeTime: str | None = None
    status: str | None = None
    trigger: str | None = None
    updateScheduledTask: bool | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.body = _RadarrCommandBody(self.body) or {}



"""Common Models."""
# pylint: disable=invalid-name, too-many-instance-attributes
from __future__ import annotations

from dataclasses import dataclass

from aiopyarr.models.base import APIResponseType, BaseModel


@dataclass(init=False)
class Diskspace(BaseModel):
    """Radarr diskspace attributes."""

    _responsetype = APIResponseType.LIST

    freeSpace: int | None = None
    label: str | None = None
    path: str | None = None
    totalSpace: int | None = None


@dataclass(init=False)
class _CommonAttrs(BaseModel):
    """Common attributes."""

    audioBitrate: int | None = None
    audioChannels: float | None = None
    audioCodec: str | None = None
    audioLanguages: str | None = None
    audioStreamCount: int | None = None
    resolution: str | None = None
    runTime: str | None = None
    scanType: str | None = None
    subtitles: str | None = None
    videoBitDepth: int | None = None
    videoBitrate: int | None = None
    videoCodec: str | None = None
    videoFps: float | None = None



"""PyArr base model."""
from __future__ import annotations

import json
from enum import Enum
from typing import Any

from ..const import LOGGER
from .const import CONVERT_TO_BOOL


class ApiJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder."""

    def default(self, o):
        """Encode default JSON."""
        if isinstance(o, BaseModel):

            return {
                key: value
                for key, value in o.__dict__.items()
                if not key.startswith("_")
            }
        if isinstance(o, Enum):
            return o.name
        return json.JSONEncoder.default(self, o)


class APIResponseType(str, Enum):
    """ApiResponseType."""

    DICT = "dict"
    LIST = "list"


class BaseModel:
    """BaseModel."""

    _datatype: BaseModel | None = None
    _responsetype: APIResponseType = APIResponseType.DICT

    def __init__(
        self, data: dict[str, Any] | list[dict[str, Any]], datatype: BaseModel = None
    ) -> None:
        """Init."""
        self._datatype = datatype
        if isinstance(data, dict):
            for key, value in data.items():
                if hasattr(self, key):
                    if hasattr(self, f"_generate_{key}"):
                        value = self.__getattribute__(f"_generate_{key}")(value)
                    self.__setattr__(key, value)

            self.__post_init__()

    def __repr__(self) -> str:
        """Representation."""
        attrs = [
            f"{key}={value}"
            for key, value in self.attributes.items()
            if value is not None and "token" not in key
        ]
        return f"{self.__class__.__name__}({', '.join(attrs)})"

    def __post_init__(self):
        """Post init."""
        for key in CONVERT_TO_BOOL:
            if hasattr(self, key) and self.__getattribute__(key) is not None:
                self.__setattr__(key, bool(self.__getattribute__(key)))
            if hasattr(self, "completionMessage"):
                if self.__getattribute__("isNewMovie") is None:
                    self.__setattr__("isNewMovie", False)
                else:
                    LOGGER.debug("isNewMovie is now always included by API")

    @property
    def attributes(self) -> dict[str, Any]:
        """Return the class attributes."""
        return {
            key: json.dumps(self.__dict__[_], cls=ApiJSONEncoder)
            for key, _ in self.__dict__.items()
            if not key.startswith("_")
        }



"""Sonarr API."""
from __future__ import annotations

from datetime import datetime
from typing import Any
from urllib.parse import quote

from aiohttp.client import ClientSession

from aiopyarr.const import HTTPMethod, HTTPResponse
from aiopyarr.exceptions import ArrInvalidCommand, ArrResourceNotFound

from .decorator import api_command
from .models.common import Diskspace
from .models.host_configuration import PyArrHostConfiguration
from .request_client import RequestClient

from .models.sonarr import (  # isort:skip
    SonarrCalendar,
    SonarrCommand,
    SonarrEpisode,
    SonarrEpisodeFile,
    SonarrEpisodeFileQuailty,
    SonarrHistory,
    SonarrParse,
    SonarrQualityProfile,
    SonarrQueueNew,
    SonarrRelease,
    SonarrRootFolder,
    SonarrSeries,
    SonarrSeriesLookup,
    SonarrSeriesUpdateParams,
    SonarrSystemBackup,
    SonarrSystemStatus,
    SonarrTag,
    SonarrWantedMissing,
)


class SonarrClient(RequestClient):  # pylint: disable=too-many-public-methods
    """API client for Sonarr endpoints."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host_configuration: PyArrHostConfiguration | None = None,
        session: ClientSession | None = None,
        hostname: str | None = None,
        ipaddress: str | None = None,
        url: str | None = None,
        api_token: str | None = None,
        port: int = 8989,
        ssl: bool | None = None,
        verify_ssl: bool | None = None,
        base_api_path: str | None = None,
        request_timeout: float = 10,
        raw_response: bool = False,
        redact: bool = True,
    ) -> None:
        """Initialize Sonarr API."""
        super().__init__(
            host_configuration,
            session,
            hostname,
            ipaddress,
            url,
            api_token,
            port,
            ssl,
            verify_ssl,
            base_api_path,
            request_timeout,
            raw_response,
            redact,
        )

    @api_command("diskspace", datatype=Diskspace)
    async def async_get_diskspace(self) -> list[Diskspace]:
        """Get information about diskspace."""

    async def async_get_episode_files(
        self, seriesid: int, episodeid: int | None = None
    ) -> SonarrEpisodeFile | list[SonarrEpisodeFile]:
        """Get information about episode files from series.

        episodeId: Get episode files matching series. Leave blank for all episodes.
        """
        return await self._async_request(
            f"episodefile{f'/{episodeid}' if episodeid is not None else ''}",
            params={"seriesId": seriesid},
            datatype=SonarrEpisodeFile,
        )

    async def async_get_image(
        self, imageid: int, imagetype: str = "poster", size: str = "large"
    ) -> bytes:
        """Get image from Sonarr.

        type: poster, banner, or fanart
        size: large, medium, small
              pixel size: None for full size (500, 250), (70, 35), (360, 180)
        """
        value = [
            value * factor
            for key, factor in {"small": 1, "medium": 2, "large": 0}.items()
            if key == size
            for key, value in {"poster": 250, "banner": 35, "fanart": 180}.items()
            if imagetype == key
        ]

        cmd = f"MediaCover/{imageid}/{imagetype}{f'-{value}' if size != 'large' else ''}.jpg"
        response = await self._async_request(cmd, datatype=SonarrCommand)
        if isinstance(response, dict):
            raise ArrResourceNotFound
        return response

    @api_command("queue", datatype=SonarrQueueNew)
    async def async_get_queue(self) -> list[SonarrQueueNew]:
        """Get information about download queue."""

    async def async_delete_queue(
        self, queueid: int, blacklist: bool = False
    ) -> HTTPResponse:
        """Delete an item from the queue and download client."""
        return await self._async_request(
            "queue",
            params={"id": queueid, "blacklist": blacklist},
            method=HTTPMethod.DELETE,
        )

    @api_command("profile", datatype=SonarrQualityProfile)
    async def async_get_profiles(self) -> list[SonarrQualityProfile]:
        """Get information about configured quality profiles."""

    @api_command("rootfolder", datatype=SonarrRootFolder)
    async def async_get_root_folders(self) -> list[SonarrRootFolder]:
        """Get information about root folders."""

    @api_command("system/status", datatype=SonarrSystemStatus)
    async def async_get_system_status(self) -> SonarrSystemStatus:
        """Get information about system status."""

    @api_command("system/backup", datatype=SonarrSystemBackup)
    async def async_get_system_backup(self) -> list[SonarrSystemBackup]:
        """Get information about system backup."""

    async def async_get_calendar(
        self, start_date: datetime | None = None, end_date: datetime | None = None
    ) -> list[SonarrCalendar]:
        """Get upcoming episodes, if start/end are not supplied episodes airing today.

        and tomorrow will be returned
        """
        params = {}
        if start_date:
            params["start"] = datetime.strptime(str(start_date), "%Y-%m-%d").strftime(
                "%Y-%m-%d"
            )
        if end_date:
            params["end"] = datetime.strptime(str(end_date), "%Y-%m-%d").strftime(
                "%Y-%m-%d"
            )

        return await self._async_request(
            "calendar", params=params, datatype=SonarrCalendar
        )

    async def async_get_commands(
        self, cmdid: int | None = None
    ) -> SonarrCommand | list[SonarrCommand]:
        """Query the status of a previously started command, or all currently started commands."""
        return await self._async_request(
            f"command{f'/{cmdid}' if cmdid is not None else ''}",
            datatype=SonarrCommand,
        )

    async def async_command_refresh_series(
        self, seriesid: int | None = None
    ) -> SonarrCommand:
        """Send refresh series command."""
        data = {"name": "RefreshSeries"}
        if seriesid is not None:
            data["seriesId"] = str(seriesid)
        return await self._async_request(
            "command",
            data=data,
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_rescan_series(
        self, seriesid: int | None = None
    ) -> SonarrCommand:
        """Send rescan series command."""
        data = {"name": "RefreshSeries"}
        if seriesid is not None:
            data["seriesId"] = str(seriesid)
        return await self._async_request(
            "command",
            data=data,
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_episode_search(self, episodeids: int) -> SonarrCommand:
        """Send episode search command."""
        return await self._async_request(
            "command",
            data={"name": "EpisodeSearch", "episodeIds": episodeids},
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_season_search(
        self, seriesid: int, seasonnumber: int
    ) -> SonarrCommand:
        """Send season search command."""
        return await self._async_request(
            "command",
            data={
                "name": "SeasonSearch",
                "seriesId": seriesid,
                "seasonNumber": seasonnumber,
            },
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_series_search(self, seriesid: int) -> SonarrCommand:
        """Send series search command."""
        return await self._async_request(
            "command",
            data={"name": "SeriesSearch", "seriesId": seriesid},
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_downloaded_episode_scan(
        self,
        path: str,
        clientid: str | None = None,
        importmode: str = "Copy",
    ) -> SonarrCommand:
        """Send series search command.

        importMode: Copy or Hardlink depending on Sonarr configuration
        """
        data = {"name": "DownloadedEpisodesScan", "path": path}
        data["importMode"] = importmode
        if clientid is not None:
            data["downloadClientId"] = clientid
        return await self._async_request(
            "command",
            data=data,
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_rss_sync(self) -> SonarrCommand:
        """Send rss sync command."""
        return await self._async_request(
            "command",
            data={"name": "RssSync"},
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_rename_files(self, files: list[int]) -> SonarrCommand:
        """Send rename files command."""
        return await self._async_request(
            "command",
            data={"name": "RenameFiles", "files": files},
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_rename_series(self, seriesids: list[int]) -> SonarrCommand:
        """Send rename series command."""
        return await self._async_request(
            "command",
            data={"name": "RenameSeries", "seriesIds": seriesids},
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_backup(self) -> SonarrCommand:
        """Send backup command."""
        return await self._async_request(
            "command",
            data={"name": "Backup"},
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_command_missing_search(self) -> SonarrCommand:
        """Send missing episode search command."""
        return await self._async_request(
            "command",
            data={"name": "missingEpisodeSearch"},
            datatype=SonarrCommand,
            method=HTTPMethod.POST,
        )

    async def async_get_episodes_by_series_id(self, seriesid: int) -> SonarrEpisode:
        """Get all episodes from a given series ID."""
        return await self._async_request("episode", params={"seriesId": seriesid})

    async def async_get_episode_by_id(self, episodeid: int) -> SonarrEpisode:
        """Get a specific episode by database id."""
        return await self._async_request(f"episode/{episodeid}", datatype=SonarrEpisode)

    async def async_update_episode(self, data: SonarrEpisode) -> SonarrEpisode:
        """Update given episodes, currently only monitored is changed."""
        return await self._async_request("episode", data=data, method=HTTPMethod.PUT)

    async def async_update_episode_file_quality(
        self, fileid: int, data: SonarrEpisodeFileQuailty
    ) -> SonarrEpisodeFile:
        """Update the quality of the episode file and returns the episode file."""
        return await self._async_request(
            f"episodefile/{fileid}", data=data, method=HTTPMethod.PUT
        )

    async def async_delete_episode_file(self, fileid: int) -> dict:
        """Delete the episode file with corresponding id.

        Args:
            id: Database id for episode file
        """
        return await self._async_request(
            f"episodefile/{fileid}", method=HTTPMethod.DELETE
        )

    async def async_get_history(  # pylint: disable=too-many-arguments
        self,
        sort_date: bool = True,
        page: int = 1,
        page_size: int = 10,
        sort_asc: bool = False,
        recordid: int | None = None,
    ) -> SonarrHistory:
        """Get history (grabs/failures/completed).

        Args:
            sort_date: date or series.title.
            page: Page number to return.
            page_size: Number of items per page.
            sort_asc: Sort items in ascending order.
            id: Filter to a specific episode ID.

        Returns:
            JSON: Array
        """
        params = {
            "page": page,
            "pageSize": page_size,
            "sortDir": "asc" if sort_asc is True else "desc",
            "sortKey": "date" if sort_date is True else "series.title",
        }
        if recordid:
            params["episodeId"] = recordid
        return await self._async_request(
            "history", datatype=SonarrHistory, params=params
        )

    async def async_get_wanted(
        self,
        sort_date: bool = True,
        page: int = 1,
        page_size: int = 10,
        sort_asc: bool = True,
    ) -> SonarrWantedMissing:
        """Get missing episode (episodes without files).

        Args:
            sort_date: airDateUtc or series.title.
            page: Page number to return.
            page_size: Number of items per page.
            sort_asc: Sort items in ascending order.
        """
        params = {
            "sortKey": "airDateUtc" if sort_date is True else "series.title",
            "page": page,
            "pageSize": page_size,
            "sortDir": "asc" if sort_asc is True else "desc",
        }
        return await self._async_request(
            "wanted/missing",
            params=params,
            datatype=SonarrWantedMissing,
        )

    async def async_parse_title_or_path(
        self, title: str | None = None, path: str | None = None
    ) -> SonarrParse:
        """Return the result of parsing a title or path.

        series and episodes will be
        returned only if the parsing matches to a specific series and one or more
        episodes. series and episodes will be formatted the same as Series and Episode
        responses.

        title: Title of series / episode
        path: file path of series / episode
        """
        if title is None and path is None:
            raise ArrInvalidCommand("A title or path must be specified")
        params = {}
        if title is not None:
            params["title"] = title
        if path is not None:
            params["path"] = path
        return await self._async_request("parse", params=params, datatype=SonarrParse)

    async def async_get_release(self, releaseid: int) -> list[SonarrRelease]:
        """Get a release with a specific episode ID."""
        return await self._async_request(
            "release",
            params={"episodeId": releaseid},
            datatype=SonarrRelease,
        )

    async def async_download_release(
        self, guid: str, indexerid: int
    ) -> list[SonarrRelease]:
        """Add a previously searched release to the download client.

        If the release is
        still in Sonarr's search cache (30 minute cache). If the release is not found
        in the cache Sonarr will return a 404.

        guid: Recently searched result guid
        """
        return await self._async_request(
            "release",
            data={"guid": guid, "indexerId": indexerid},
            datatype=SonarrRelease,
            method=HTTPMethod.POST,
        )

    async def async_push_release(
        self,
        title: str,
        download_url: str,
        protocol: str,
        publish_date: str,
    ) -> list[SonarrRelease]:
        """If the title is wanted, Sonarr will grab it.

        Args:
            title: Release name
            download_url: .torrent file URL
            protocol: "Usenet" or "Torrent
            publish_date: ISO8601 date string
        """
        data = {
            "title": title,
            "downloadUrl": download_url,
            "protocol": protocol,
            "publishDate": publish_date,
        }
        return await self._async_request(
            "release/push",
            data=data,
            datatype=SonarrRelease,
            method=HTTPMethod.POST,
        )

    async def async_get_series(
        self, seriesid: int | None = None
    ) -> SonarrSeries | list[SonarrSeries]:
        """Return all series in your collection or the series with the matching.

        series ID if one is found.
        """
        return await self._async_request(
            f"series{f'/{seriesid}' if seriesid is not None else ''}",
            datatype=SonarrSeries,
        )

    async def async_get_series_episodes(
        self, seriesid: int, episodeid: int | None = None
    ) -> list[SonarrEpisode]:
        """Get information about episodes from series."""
        response = await self._async_request(
            f"episode{f'/{episodeid}' if episodeid is not None else ''}",
            params={"seriesId": seriesid},
            datatype=SonarrEpisode,
        )
        if "message" in response and response["message"] == "NotFound":
            raise ArrResourceNotFound
        if "seriesId" in response and response["seriesId"] != seriesid:
            raise ArrResourceNotFound
        return response

    async def async_add_series(  # pylint: disable=too-many-arguments
        self,
        tvdb_id: int,
        quality_profile_id: int,
        root_dir: str,
        season_folder: bool = True,
        monitored: bool = True,
        ignore_episodes_with_files: bool = False,
        ignore_episodes_without_files: bool = False,
        search_for_missing_episodes: bool = False,
    ) -> SonarrSeries:
        """Add a new series to your collection.

        Args:
            tvdb_id: TDDB Id
            quality_profile_id: Database id for quality profile
            root_dir: Root folder location, full path will be created from this
            season_folder: Create a folder for each season.
            monitored: Monitor this series.
            ignore_episodes_with_files: Ignore any episodes with existing files.
            ignore_episodes_without_files: Ignore any episodes without existing files.
            search_for_missing_episodes: Search for missing episodes to download.
        """
        series_json = await self._async_construct_series_json(
            tvdb_id,
            quality_profile_id,
            root_dir,
            season_folder,
            monitored,
            ignore_episodes_with_files,
            ignore_episodes_without_files,
            search_for_missing_episodes,
        )
        return await self._async_request(
            "series",
            data=series_json,
            datatype=SonarrSeries,
            method=HTTPMethod.POST,
        )

    async def async_update_series(self, data: SonarrSeriesUpdateParams) -> SonarrSeries:
        """Update an existing series."""
        return await self._async_request(
            "series",
            data=data,
            datatype=SonarrSeries,
            method=HTTPMethod.PUT,
        )

    async def async_delete_series(
        self, seriesid: int, delete_files: bool = False
    ) -> dict:
        """Delete the series with the given id.

        delete_files: If true series folder and files will be deleted.
        """
        return await self._async_request(
            f"series/{seriesid}",
            params={"id": seriesid, "deleteFiles": str(delete_files)},
            method=HTTPMethod.DELETE,
        )

    @api_command("series/lookup", datatype=SonarrSeriesLookup)
    async def async_get_series_lookup(self) -> list[SonarrSeriesLookup]:
        """Lookup information about a series."""

    async def async_lookup_series(
        self, term: str | None = None, seriesid: int | None = None
    ) -> list[SonarrSeriesLookup]:
        """Search for new shows on TheTVDB.com utilizing sonarr.tv's.

        caching and augmentation proxy.
        """
        if term is None and seriesid is None:
            raise ArrResourceNotFound("A term or TVDB id must be included")
        return await self._async_request(
            "series/lookup",
            datatype=SonarrSeriesLookup,
            params={"term": quote(term, safe="") if term else f"tvdb:{seriesid}"},
        )

    async def async_get_tags(
        self, tagid: int | None = None
    ) -> SonarrTag | list[SonarrTag]:
        """Return all tags or specific tag by database id.

        id: Get tag matching id. Leave blank for all.
        """
        return await self._async_request(
            f"tag{f'/{tagid}' if tagid is not None else ''}",
            datatype=SonarrTag,
        )

    async def async_create_tag(self, label: str) -> SonarrTag:
        """Add a new tag.

        label: Tag name / label
        """
        return await self._async_request(
            "tag",
            data={"label": label},
            datatype=SonarrTag,
            method=HTTPMethod.POST,
        )

    async def async_update_tag(self, data: SonarrTag) -> SonarrTag:
        """Update an existing tag."""
        return await self._async_request("tag", data=data, method=HTTPMethod.PUT)

    async def async_delete_tag(self, tagid: int) -> dict:
        """Delete the tag with the given ID."""
        return await self._async_request(f"tag/{tagid}", method=HTTPMethod.DELETE)

    async def _async_construct_series_json(  # pylint: disable=too-many-arguments
        self,
        tvdb_id: int,
        quality_profile_id: int,
        root_dir: str,
        season_folder: bool = True,
        monitored: bool = True,
        ignore_episodes_with_files: bool = False,
        ignore_episodes_without_files: bool = False,
        search_for_missing_episodes: bool = False,
    ) -> dict[str, Any] | None:
        """Search for new shows on trakt and returns Series JSON to add.

        Args:
            tvdb_id: TVDB id to search
            quality_profile_id: Database id for Quality profile
            root_dir: Root directory for media
            season_folder: Specify if a season folder should be created.
            monitored: Specify if the series should be monitored.
            ignore_episodes_with_files: Ignore episodes that already have files.
            ignore_episodes_without_files: Ignore episodes that dont have any files.
            search_for_missing_episodes: Search for any missing episodes and download them.
        """
        result = await self.async_lookup_series(seriesid=tvdb_id)
        if not isinstance(result, list):
            return None
        series = result[0]
        assert isinstance(series, dict)
        if not monitored and series.get("seasons"):
            for season in series["seasons"]:
                season["monitored"] = False

        return {
            "title": series["title"],
            "seasons": series["seasons"],
            "rootFolderPath": root_dir,
            "qualityProfileId": quality_profile_id,
            "seasonFolder": season_folder,
            "monitored": monitored,
            "tvdbId": tvdb_id,
            "images": series["images"],
            "titleSlug": series["titleSlug"],
            "addOptions": {
                "ignoreEpisodesWithFiles": ignore_episodes_with_files,
                "ignoreEpisodesWithoutFiles": ignore_episodes_without_files,
                "searchForMissingEpisodes": search_for_missing_episodes,
            },
        }



"""Sonarr Common Models."""
# pylint: disable=invalid-name, too-many-instance-attributes, too-few-public-methods
from __future__ import annotations

from dataclasses import dataclass

from .base import APIResponseType, BaseModel
from .common import _CommonAttrs


@dataclass(init=False)
class _SonarrCommon6(BaseModel):
    """Sonarr common attributes."""

    id: int | None = None


@dataclass(init=False)
class _SonarrSeriesAlternateTitle(BaseModel):
    """Sonarr series alternate titles attributes."""

    seasonNumber: int | None = None
    title: str | None = None


@dataclass(init=False)
class _SonarrCommon5(_SonarrSeriesAlternateTitle, _SonarrCommon6):
    """Sonarr common attributes."""

    airDate: str | None = None
    airDateUtc: str | None = None
    episodeFileId: int | None = None
    episodeNumber: int | None = None
    hasFile: bool | None = None
    monitored: bool | None = None
    overview: str | None = None
    seriesId: int | None = None


@dataclass(init=False)
class _SonarrCommon4(_SonarrCommon6):
    """Sonarr common attributes."""

    allowed: list[_SonarrQualityProfileValueAttr] | None = None
    cutoff: _SonarrCutoff | None = None
    items: list[_SonarrQualityProfileValueItems] | None = None
    name: str | None = None

    def __post_init__(self):
        super().__post_init__()
        self.allowed = [
            _SonarrQualityProfileValueAttr(allowed) for allowed in self.allowed or []
        ]
        if isinstance(self.cutoff, dict):
            self.cutoff = _SonarrCutoff(self.cutoff) or {}
        if isinstance(self.items, list):
            self.items = [
                _SonarrQualityProfileValueItems(item) for item in self.items or []
            ]


@dataclass(init=False)
class _SonarrQualityProfileValueAttr(_SonarrCommon4):
    """Sonarr quality profile value attributes."""

    weight: int | None = None


@dataclass(init=False)
class _SonarrCutoff(_SonarrCommon6):
    """Sonarr cutoff attributes."""

    name: str | None = None
    weight: int | None = None


@dataclass(init=False)
class _SonarrQualityProfileValueItems(BaseModel):
    """Sonarr quality profile value items attributes."""

    allowed: bool | None = None
    quality: _SonarrCutoff | None = None

    def __post_init__(self):
        super().__post_init__()
        self.quality = _SonarrCutoff(self.quality) or {}


@dataclass(init=False)
class _SonarrQualityProfile(BaseModel):
    """Sonarr quality profile attributes."""

    isLoaded: bool | None = None
    value: _SonarrQualityProfileValueAttr | None = None

    def __post_init__(self):
        super().__post_init__()
        self.value = _SonarrQualityProfileValueAttr(self.value) or {}


@dataclass(init=False)
class _SonarrImages(BaseModel):
    """Sonarr images attributes."""

    coverType: str | None = None
    remoteUrl: str | None = None
    url: str | None = None


@dataclass(init=False)
class _SonarrSeriesCommon3:
    """Sonarr series common attributes."""

    images: list[_SonarrImages] | None = None
    title: str | None = None
    titleSlug: str | None = None
    tvdbId: int | None = None


@dataclass(init=False)
class _SonarrSeriesCommon2(_SonarrSeriesCommon3):
    """Sonarr series common attributes."""

    airTime: str | None = None
    cleanTitle: str | None = None
    ended: bool | None = None
    firstAired: str | None = None
    imdbId: str | None = None
    languageProfileId: int | None = None
    monitored: bool | None = None
    network: str | None = None
    qualityProfile: _SonarrQualityProfile | None = None
    qualityProfileId: int | None = None
    runtime: int | None = None
    seasonFolder: bool | None = None
    seriesType: str | None = None
    status: str | None = None
    tvRageId: int | None = None
    useSceneNumbering: bool | None = None
    year: int | None = None

    def __post_init__(self):
        self.images = [_SonarrImages(image) for image in self.images or []]
        if isinstance(self.qualityProfile, dict):
            self.qualityProfile = _SonarrQualityProfile(self.qualityProfile) or {}


@dataclass(init=False)
class _SonarrSeries2(_SonarrSeriesCommon2):
    """Sonarr series attributes."""

    lastInfoSync: str | None = None
    overview: str | None = None
    path: str | None = None
    qualityProfile: _SonarrQualityProfile | None = None


@dataclass(init=False)
class _SonarrEpisodeFileCommon:
    """Sonarr episode file common attributes."""

    languageCutoffNotMet: bool | None = None
    quality: _SonarrQualitySub | None = None
    qualityCutoffNotMet: bool | None = None
    seriesId: int | None = None

    def __post_init__(self):
        self.quality = _SonarrQualitySub(self.quality) or {}


@dataclass(init=False)
class _SonarrCommon8:
    """Sonarr episode file common attributes."""

    dateAdded: str | None = None
    path: str | None = None
    sceneName: str | None = None
    seasonNumber: int | None = None
    size: int | None = None


@dataclass(init=False)
class _SonarrEpisodeFile(_SonarrEpisodeFileCommon, _SonarrCommon6, _SonarrCommon8):
    """Sonarr episode file attributes."""

    language: _SonarrCutoff | None = None
    mediaInfo: _CommonAttrs | None = None
    relativePath: str | None = None
    releaseGroup: str | None = None

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.language, dict):
            self.language = _SonarrCutoff(self.language) or {}
        if isinstance(self.mediaInfo, dict):
            self.mediaInfo = _CommonAttrs(self.mediaInfo) or {}


@dataclass(init=False)
class _SonarrCommon2(_SonarrCommon5):
    """Sonarr common attributes."""

    absoluteEpisodeNumber: int | None = None
    episodeFile: _SonarrEpisodeFile | None = None

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.episodeFile, dict):
            self.episodeFile = _SonarrEpisodeFile(self.episodeFile) or {}


@dataclass(init=False)
class _SonarrEpisodeQuality:
    """Sonarr episode quality attributes."""

    proper: bool | None = None
    quality: _SonarrQualitySubSub | None = None

    def __post_init__(self):
        self.quality = _SonarrQualitySubSub(self.quality) or {}


@dataclass(init=False)
class _SonarrRevisionAttr(BaseModel):
    """Sonarr revision attributes."""

    real: int | None = None
    version: int | None = None
    isRepack: bool | None = None


class _SonarrQualitySubSub(_SonarrCommon6):
    """Sonarr quality attributes."""

    allowed: list | None = []
    name: str | None = None
    source: str | None = None
    resolution: int | None = None


@dataclass(init=False)
class _SonarrQualitySub(BaseModel):
    """Sonarr quality attributes."""

    quality: _SonarrQualitySubSub | None = None
    proper: bool | None = None
    revision: _SonarrRevisionAttr | None = None

    def __post_init__(self):
        super().__post_init__()
        self.quality = _SonarrQualitySubSub(self.quality) or {}
        if isinstance(self.revision, dict):
            self.revision = _SonarrRevisionAttr(self.revision) or {}


@dataclass(init=False)
class _SonarrCommon7(BaseModel):
    """Sonarr common attributes."""

    age: int | None = None
    approved: bool | None = None
    downloadAllowed: bool | None = None
    downloadUrl: str | None = None
    guid: str | None = None
    indexer: str | None = None
    indexerId: int | None = None
    rejections: list[str] | None = None
    sceneSource: bool | None = None
    seasonNumber: int | None = None
    size: int | None = None
    title: str | None = None
    tvRageId: int | None = None


@dataclass(init=False)
class _SonarrHistoryRecordData(_SonarrCommon7):
    """Sonarr history record data attributes."""

    downloadClient: str | None = None
    droppedPath: str | None = None
    importedPath: str | None = None
    nzbInfoUrl: str | None = None
    releaseGroup: str | None = None
    ageHours: float | None = None
    ageMinutes: float | None = None
    fileId: int | None = None
    downloadClientName: str | None = None
    preferredWordScore: int | None = None
    publishedDate: str | None = None
    tvdbId: str | None = None
    protocol: int | None = None
    torrentInfoHash: str | None = None


@dataclass(init=False)
class _SonarrEpisodeFile2(_SonarrCommon6, _SonarrCommon8):
    """Sonarr episode file attributes."""

    quality: _SonarrEpisodeQuality | None = None
    seriesId: int | None = None


@dataclass(init=False)
class _SonarrHistoryRecordEpisode(_SonarrCommon2):
    """Sonarr history record episode attributes."""

    unverifiedSceneNumbering: bool | None = None


@dataclass(init=False)
class _SonarrRatings(BaseModel):
    """Sonarr ratings attributes."""

    value: float | None = None
    votes: int | None = None


@dataclass(init=False)
class _SonarrSeasonStatistics(BaseModel):
    """Sonarr season statistics attributes."""

    episodeCount: int | None = None
    episodeFileCount: int | None = None
    percentOfEpisodes: float | None = None
    previousAiring: str | None = None
    sizeOnDisk: int | None = None
    totalEpisodeCount: int | None = None


@dataclass(init=False)
class _SonarrWantedMissingSeriesSeason(BaseModel):
    """Sonarr wanted missing series season attributes."""

    monitored: bool | None = None
    seasonNumber: int | None = None
    statistics: _SonarrSeasonStatistics | None = None

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.statistics, dict):
            self.statistics = _SonarrSeasonStatistics(self.statistics) or {}


@dataclass(init=False)
class _SonarrSeriesCommon(_SonarrSeriesCommon2, BaseModel):
    """Sonarr series common attributes."""

    added: str | None = None
    certification: str | None = None
    genres: list[str] | None = None
    overview: str | None = None
    profileId: int | None = None
    ratings: _SonarrRatings | None = None
    seasonCount: int | None = None
    seasons: list[_SonarrWantedMissingSeriesSeason] | None = None
    sortTitle: str | None = None
    tags: list[int] | None = None
    tvMazeId: int | None = None

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.ratings, dict):
            self.ratings = _SonarrRatings(self.ratings) or {}
        if isinstance(self.seasons, list):
            self.seasons = [
                _SonarrWantedMissingSeriesSeason(season)
                for season in self.seasons or []
            ]


@dataclass(init=False)
class _SonarrHistoryRecordSeries(_SonarrSeriesCommon, _SonarrCommon6):
    """Sonarr history record series attributes."""

    lastInfoSync: str | None = None
    path: str | None = None


@dataclass(init=False)
class _SonarrHistoryRecord(_SonarrEpisodeFileCommon, _SonarrCommon6):
    """Sonarr history record attributes."""

    data: _SonarrHistoryRecordData | None = None
    date: str | None = None
    downloadId: str | None = None
    episodeId: int | None = None
    eventType: str | None = None
    language: _SonarrCommon4 | None = None
    sourceTitle: str | None = None

    def __post_init__(self):
        super().__post_init__()
        self.data = _SonarrHistoryRecordData(self.data) or {}
        self.language = _SonarrCommon4(self.language) or {}


@dataclass(init=False)
class _SonarrCommon(BaseModel):
    """Sonarr common attributes."""

    page: int | None = None
    pageSize: int | None = None
    sortDirection: str | None = None
    sortKey: str | None = None
    totalRecords: int | None = None


@dataclass(init=False)
class _SonarrWantedMissingSeries(_SonarrSeries2, _SonarrCommon6):
    """Sonarr wanted missing series attributes."""

    seasons: list[_SonarrWantedMissingSeriesSeason] | None = None

    def __post_init__(self):
        super().__post_init__()
        self.seasons = [
            _SonarrWantedMissingSeriesSeason(season) for season in self.seasons or []
        ]


@dataclass(init=False)
class _SonarrWantedMissingRecord(_SonarrCommon2):
    """Sonarr wanted missing record attributes."""

    downloading: bool | None = None
    sceneEpisodeNumber: int | None = None
    sceneSeasonNumber: int | None = None
    series: _SonarrWantedMissingSeries | None = None
    tvDbEpisodeId: int | None = None

    def __post_init__(self):
        super().__post_init__()
        self.series = _SonarrWantedMissingSeries(self.series) or {}


@dataclass(init=False)
class _SonarrSeriesTitleInfo(BaseModel):
    """Sonarr series title info attributes."""

    title: str | None = None
    titleWithoutYear: str | None = None
    year: int | None = None


@dataclass(init=False)
class _SonarrCommon3(BaseModel):
    """Sonarr common attributes."""

    _responsetype = APIResponseType.LIST

    episodeNumbers: list[int] | None = None
    fullSeason: bool | None = None
    language: str | None = None
    quality: _SonarrQualitySub | None = None
    releaseGroup: str | None = None
    seriesTitle: str | None = None

    def __post_init__(self):
        super().__post_init__()
        self.quality = _SonarrQualitySub(self.quality) or {}


@dataclass(init=False)
class _SonarrParseEpisodeInfo(_SonarrCommon3):
    """Sonarr parse episode info attributes."""

    absoluteEpisodeNumbers: list[int] | None = None
    isAbsoluteNumbering: bool | None = None
    isDaily: bool | None = None
    isPossibleSpecialEpisode: bool | None = None
    releaseHash: str | None = None
    releaseTitle: str | None = None
    seasonNumber: int | None = None
    seriesTitleInfo: _SonarrSeriesTitleInfo | None = None
    special: bool | None = None

    def __post_init__(self):
        super().__post_init__()
        self.seriesTitleInfo = _SonarrSeriesTitleInfo(self.seriesTitleInfo) or {}


@dataclass(init=False)
class _SonarrLogRecord(_SonarrCommon6):
    """Sonarr log record attributes."""

    time: str | None = None
    level: str | None = None
    logger: str | None = None
    message: str | None = None


@dataclass(init=False)
class _FileRevision:
    """File revision for updating episode file."""

    version: int | None = None
    real: int | None = None


@dataclass(init=False)
class _FileEpisodeQuality:
    """File quality for updating episode file."""

    quality: _SonarrCommon6 | None = None
    revision: _FileRevision | None = None



"""Radarr API."""
from __future__ import annotations

from datetime import datetime

from aiohttp.client import ClientSession

from .const import HTTPMethod, HTTPResponse
from .decorator import api_command
from .models.common import Diskspace
from .models.host_configuration import PyArrHostConfiguration
from .request_client import RequestClient

from .models.radarr import (  # isort:skip
    RadarrBlocklist,
    RadarrBlocklistMovie,
    RadarrCalendar,
    RadarrCommand,
    RadarrCustomFilter,
    RadarrDownloadClient,
    RadarrHealth,
    RadarrHostConfig,
    RadarrImportList,
    RadarrIndexer,
    RadarrMetadataConfig,
    RadarrMovie,
    RadarrMovieEditor,
    RadarrMovieFile,
    RadarrMovieHistory,
    RadarrNamingConfig,
    RadarrNotification,
    RadarrQualityProfile,
    RadarrQueue,
    RadarrQueueDetail,
    RadarrQueueStatus,
    RadarrRemotePathMapping,
    RadarrRootFolder,
    RadarrSystemStatus,
    RadarrTag,
    RadarrTagDetails,
    RadarrUIConfig,
    RadarrUpdate,
)


class RadarrClient(RequestClient):  # pylint: disable=too-many-public-methods
    """API client for Radarr endpoints."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host_configuration: PyArrHostConfiguration | None = None,
        session: ClientSession | None = None,
        hostname: str | None = None,
        ipaddress: str | None = None,
        url: str | None = None,
        api_token: str | None = None,
        port: int = 7878,
        ssl: bool | None = None,
        verify_ssl: bool | None = None,
        base_api_path: str | None = None,
        request_timeout: float = 10,
        raw_response: bool = False,
        redact: bool = True,
    ) -> None:
        """Initialize Radarr API."""
        super().__init__(
            host_configuration,
            session,
            hostname,
            ipaddress,
            url,
            api_token,
            port,
            ssl,
            verify_ssl,
            base_api_path,
            request_timeout,
            raw_response,
            redact,
        )

    async def async_get_movies(
        self, movieid: int | None = None
    ) -> RadarrMovie | list[RadarrMovie]:
        """Get information about movies.

        Include a TMDB id for a specific movie or leave black for all.
        """
        return await self._async_request(
            "movie",
            params={"tmdbId": movieid} if movieid is not None else None,
            datatype=RadarrMovie,
        )

    async def async_add_movie(  # pylint: disable=too-many-arguments
        self,
        db_id: str,
        quality_profile_id: int,
        root_dir: str,
        monitored: bool = True,
        search_for_movie: bool = True,
        tmdb: bool = True,
    ) -> HTTPResponse:
        """Add a movie to the database.

        Args:
            db_id: IMDB or TMDB ID
            quality_profile_id: ID of the quality profile the movie will use
            root_dir: location of the root DIR
            monitored: should the movie be monitored.
            search_for_movie: Should we search for the movie.
            tmdb: Use TMDB IDs. Set to False to use IMDB.
        """
        movie_json = await self._async_construct_movie_json(
            db_id, quality_profile_id, root_dir, monitored, search_for_movie, tmdb
        )
        return await self._async_request(
            "movie",
            params=movie_json,
            datatype=RadarrMovie,
            method=HTTPMethod.POST,
        )

    async def async_update_movies(
        self, data: RadarrMovie | RadarrMovieEditor, move_files: bool = False
    ) -> RadarrMovie | list[RadarrMovie]:
        """Edit movie properties of multiple movies at once."""
        return await self._async_request(
            "movie/editor" if hasattr(data, "movieIds") else "movie",
            params=None if hasattr(data, "movieIds") else {"moveFiles": move_files},
            data=data,
            datatype=RadarrMovieEditor,
            method=HTTPMethod.PUT,
        )

    async def async_delete_movies(
        self,
        ids: int | list[int],
        delete_files: bool = False,
        add_exclusion: bool = False,
    ) -> HTTPResponse:
        """Delete movies (and optionally files).

        ids: include an integer to delete one movie or a list for mass deletion
        """
        data = {
            "movieIds": ids,
            "deleteFiles": delete_files,
            "addImportExclusion": add_exclusion,
        }
        return await self._async_request(
            "movie/editor" if isinstance(ids, list) else f"movie/{ids}",
            data=data,
            method=HTTPMethod.DELETE,
        )

    async def async_import_movies(self, data: list[RadarrMovie]) -> list[RadarrMovie]:
        """Import movies in bulk.

        It allows movies to be bulk added to the Radarr database.
        """
        return await self._async_request(
            "movie/import",
            data=data,
            datatype=RadarrMovie,
            method=HTTPMethod.POST,
        )

    async def async_get_movie_files_by_movie_id(self, movieid: int) -> RadarrMovieFile:
        """Get a movie file object by Movie database id."""
        return await self._async_request(
            "moviefile",
            params={"movieid": movieid},
            datatype=RadarrMovieFile,
        )

    async def async_delete_movie_file(self, movieid: int) -> HTTPResponse:
        """Delete a moviefile by its database id."""
        return await self._async_request(
            f"moviefile/{movieid}",
            method=HTTPMethod.DELETE,
        )

    async def async_lookup_movie(
        self, term: str, tmdb: bool = True
    ) -> list[RadarrMovie]:
        """Lookup information about movie.

        tmdb: Use TMDB IDs. Set to False to use IMDB.
        """
        return await self._async_request(
            "movie/lookup",
            params={"term": f"{'tmdb' if tmdb is True else 'imdb'} :{term}"},
            datatype=RadarrMovie,
        )

    async def async_lookup_movie_files(self, ids: list[int]) -> list[RadarrMovieFile]:
        """Get movie file information for multiple movie files."""
        return await self._async_request(
            "moviefile", params={"moviefileids": ids}, datatype=RadarrMovieFile
        )

    async def async_get_history(
        self,
        page: int = 1,
        page_size: int = 20,
        sort_direction: str = "descending",
        sort_key: str = "date",
    ) -> list[RadarrMovieHistory]:
        """Return a json object list of items in your history.

        Args:
            page: Page to be returned.
            page_size: Number of results per page.
            sort_direction: Direction to sort items.
            sort_key: Field to sort by.
        """
        params = {
            "page": page,
            "pageSize": page_size,
            "sortDirection": sort_direction,
            "sortKey": sort_key,
        }
        return await self._async_request(
            "history",
            params=params,
            datatype=RadarrMovieHistory,
        )

    async def async_get_movie_history(
        self, recordid: int, event_type: int | None = None
    ) -> list[RadarrMovieHistory]:
        """Get history for a given movie in database by its database id.

        Args:
            id: Database id of movie
            event_type: History event type to retrieve.
        """
        params = {"movieId": recordid}
        if event_type:
            params["eventType"] = event_type
        return await self._async_request(
            "history/movie",
            params=params,
            datatype=RadarrMovieHistory,
        )

    async def async_get_download_clients(
        self, clientid: int | None = None
    ) -> RadarrDownloadClient | list[RadarrDownloadClient]:
        """Get information about download client.

        id: Get downloadclient matching id. Leave blank for all.
        """
        return await self._async_request(
            f"downloadclient{f'/{clientid}' if clientid is not None else ''}",
            datatype=RadarrDownloadClient,
        )

    async def async_get_import_lists(
        self, listid: int | None = None
    ) -> RadarrImportList | list[RadarrImportList]:
        """Get information about import list.

        id: Get import list matching id. Leave blank for all.
        """
        return await self._async_request(
            f"importlist{f'/{listid}' if listid is not None else ''}",
            datatype=RadarrImportList,
        )

    @api_command("diskspace", datatype=Diskspace)
    async def async_get_diskspace(self) -> list[Diskspace]:
        """Get information about diskspace."""

    @api_command("config/ui", datatype=RadarrUIConfig)
    async def async_get_ui_config(self) -> RadarrUIConfig:
        """Get information about UI configuration."""

    async def async_update_ui_config(self, data: RadarrUIConfig) -> HTTPResponse:
        """Edit one or many UI settings and save to to the database."""
        return await self._async_request(
            "config/ui",
            data=data,
            datatype=RadarrUIConfig,
            method=HTTPMethod.PUT,
        )

    @api_command("config/host", datatype=RadarrHostConfig)
    async def async_get_host_config(self) -> RadarrHostConfig:
        """Get information about host configuration."""

    async def async_update_host_config(self, data: RadarrHostConfig) -> HTTPResponse:
        """Edit General/Host settings for Radarr."""
        return await self._async_request(
            "config/host",
            data=data,
            datatype=RadarrHostConfig,
            method=HTTPMethod.PUT,
        )

    @api_command("config/naming", datatype=RadarrNamingConfig)
    async def async_get_naming_config(self) -> RadarrNamingConfig:
        """Get information about naming configuration."""

    async def async_update_naming_config(
        self, data: RadarrNamingConfig
    ) -> HTTPResponse:
        """Edit Settings for movie file and folder naming."""
        return await self._async_request(
            "config/naming",
            data=data,
            datatype=RadarrNamingConfig,
            method=HTTPMethod.PUT,
        )

    @api_command("metadata", datatype=RadarrMetadataConfig)
    async def async_get_metadata_config(self) -> RadarrMetadataConfig:
        """Get information about metadata configuration."""

    async def async_get_tags_details(
        self, tagid: int | None = None
    ) -> RadarrTagDetails | list[RadarrTagDetails]:
        """Get information about tag details.

        id: Get tag details matching id. Leave blank for all.
        """
        return await self._async_request(
            f"tag/detail{f'/{tagid}' if tagid is not None else ''}",
            datatype=RadarrTagDetails,
        )

    async def async_get_tags(
        self, tagid: int | None = None
    ) -> RadarrTag | list[RadarrTag]:
        """Get information about tag.

        id: Get tag matching id. Leave blank for all.
        """
        return await self._async_request(
            f"tag{f'/{tagid}' if tagid is not None else ''}",
            datatype=RadarrTag,
        )

    async def async_get_blocklist(
        self,
        page: int = 1,
        page_size: int = 20,
        sort_direction: str = "descending",
        sort_key: str = "date",
    ) -> RadarrBlocklist:
        """Return blocklisted releases.

        Args:
            page: Page to be returned.
            page_size: Number of results per page.
            sort_direction: Direction to sort items.
            sort_key: Field to sort by.
        """
        params = {
            "page": page,
            "pageSize": page_size,
            "sortDirection": sort_direction,
            "sortKey": sort_key,
        }
        return await self._async_request(
            "blocklist",
            params=params,
            datatype=RadarrBlocklist,
        )

    async def async_delete_blocklist(self, bocklistid: int) -> HTTPResponse:
        """Remove a specific release (the id provided) from the blocklist."""
        return await self._async_request(
            "blocklist",
            params={"id": bocklistid},
            method=HTTPMethod.DELETE,
        )

    async def async_get_blocklist_movie(
        self,
        bocklistid: int,
    ) -> list[RadarrBlocklistMovie]:
        """Retrieve blocklisted releases that are tied to a given movie in the database."""
        return await self._async_request(
            "blocklist/movie",
            params={"movieId": bocklistid},
            datatype=RadarrBlocklistMovie,
        )

    async def async_delete_blocklist_bulk(self, data: list[int]) -> HTTPResponse:
        """Delete blocklisted releases in bulk."""
        return await self._async_request(
            "blocklist/bulk",
            data=data,
            method=HTTPMethod.DELETE,
        )

    @api_command("queue/status", datatype=RadarrQueueStatus)
    async def async_get_queue_status(self) -> RadarrQueueStatus:
        """Get information about download queue status."""

    async def async_get_queue(  # pylint: disable=too-many-arguments
        self,
        page: int = 1,
        page_size: int = 20,
        sort_direction: str = "ascending",
        sort_key: str = "timeLeft",
        include_unknown_movie_items: bool = True,
    ) -> list[RadarrQueue]:
        """Return a json object list of items in the queue.

        Args:
            page: Page to be returned.
            page_size: Number of results per page.
            sort_direction: Direction to sort items.
            sort_key: Field to sort by.
            include_unknown_movie_items: Include unknown movie items.
        """
        params = {
            "page": page,
            "pageSize": page_size,
            "sortDirection": sort_direction,
            "sortKey": sort_key,
            "includeUnknownMovieItems": include_unknown_movie_items,
        }
        return await self._async_request("queue", params=params, datatype=RadarrQueue)

    async def async_get_queue_details(
        self,
        include_movie: bool = True,
    ) -> list[RadarrQueueDetail]:
        """Get details of all items in queue."""
        return await self._async_request(
            "queue/details",
            params={"includeMovie": str(include_movie)},
            datatype=RadarrQueueDetail,
        )

    async def async_force_grab_queue_item(self, queueid: int) -> HTTPResponse:
        """Perform a Radarr "force grab" on a pending queue item by its ID."""
        return await self._async_request(
            f"queue/grab/{queueid}", method=HTTPMethod.POST
        )

    async def async_delete_queue(
        self,
        ids: int | list[int],
        remove_from_client: bool = True,
        blocklist: bool = True,
    ) -> HTTPResponse:
        """Remove an item from the queue and optionally blocklist it.

        Args:
            ids: id of the item to be removed or mass deletion with a list
            remove_from_client: Remove the item from the client.
            blocklist: Add the item to the blocklist.
        """
        return await self._async_request(
            "queue/bulk" if isinstance(ids, list) else f"queue/{ids}",
            params={"removeFromClient": remove_from_client, "blocklist": blocklist},
            data=ids if isinstance(ids, list) else None,
            method=HTTPMethod.DELETE,
        )

    async def async_get_indexers(
        self, indexerid: int | None = None
    ) -> RadarrIndexer | list[RadarrIndexer]:
        """Get all indexers or a single indexer by its database id.

        id: Get indexer matching id. Leave blank for all.
        """
        return await self._async_request(
            f"indexer{f'/{indexerid}' if indexerid is not None else ''}",
            datatype=RadarrIndexer,
        )

    async def async_update_indexer(
        self, indexerid: int, data: RadarrIndexer
    ) -> HTTPResponse:
        """Edit an indexer."""
        return await self._async_request(
            f"indexer/{indexerid}",
            data=data,
            method=HTTPMethod.PUT,
        )

    async def async_delete_indexer(self, indexerid: int) -> HTTPResponse:
        """Delete indexer by database id."""
        return await self._async_request(
            f"indexer/{indexerid}", method=HTTPMethod.DELETE
        )

    async def async_update_download_client(
        self, clientid: int, data: RadarrDownloadClient
    ) -> HTTPResponse:
        """Edit a download client by database id.

        Args:
            id: Download client database id
            data: data to be updated within download client
        """
        return await self._async_request(
            f"downloadclient/{clientid}",
            data=data,
            datatype=RadarrDownloadClient,
            method=HTTPMethod.PUT,
        )

    async def async_delete_download_client(self, clientid: int) -> HTTPResponse:
        """Delete a download client."""
        return await self._async_request(
            f"downloadclient/{clientid}", method=HTTPMethod.DELETE
        )

    async def async_update_import_list(
        self, listid: int, data: RadarrImportList
    ) -> HTTPResponse:
        """Edit an importlist."""
        return await self._async_request(
            f"importlist/{listid}",
            data=data,
            datatype=RadarrImportList,
            method=HTTPMethod.PUT,
        )

    async def async_delete_import_list(self, listid: int) -> HTTPResponse:
        """Delete an import list."""
        return await self._async_request(
            f"importlist/{listid}",
            datatype=RadarrIndexer,
            method=HTTPMethod.DELETE,
        )

    async def async_get_notifications(
        self, notifyid: int | None = None
    ) -> RadarrNotification | list[RadarrNotification]:
        """Get information about notification.

        id: Get notification matching id. Leave blank for all.
        """
        return await self._async_request(
            f"notification{f'/{notifyid}' if notifyid is not None else ''}",
            datatype=RadarrNotification,
        )

    async def async_update_notification(
        self, notifyid: int, data: RadarrNotification
    ) -> HTTPResponse:
        """Edit a notification."""
        return await self._async_request(
            f"notification/{notifyid}",
            data=data,
            datatype=RadarrNotification,
            method=HTTPMethod.PUT,
        )

    async def async_delete_notification(self, notifyid: int) -> HTTPResponse:
        """Delete a notification."""
        return await self._async_request(
            f"notification/{notifyid}",
            datatype=RadarrNotification,
            method=HTTPMethod.DELETE,
        )

    async def async_create_tag(self, label: str) -> HTTPResponse:
        """Create a new tag.

        Can be assigned to a movie, list, delay profile, notification, or restriction.
        """
        return await self._async_request(
            "tag",
            data={"id": 0, "label": label},
            datatype=RadarrTag,
            method=HTTPMethod.POST,
        )

    async def async_update_tag(self, tagid: int, label: str) -> HTTPResponse:
        """Edit a tag by its database id."""
        return await self._async_request(
            f"tag/{tagid}",
            data={"id": tagid, "label": label},
            datatype=RadarrTag,
            method=HTTPMethod.PUT,
        )

    async def async_delete_tag(self, tagid: int) -> HTTPResponse:
        """Delete a tag."""
        return await self._async_request(
            f"tag/{tagid}",
            datatype=RadarrTag,
            method=HTTPMethod.DELETE,
        )

    @api_command("command", datatype=RadarrCommand)
    async def async_get_command(self) -> list[RadarrCommand]:
        """Get status of recently run commands/tasks."""

    async def async_command_app_update(self) -> HTTPResponse:
        """Trigger Radarr software update."""
        return await self._async_request(
            "command",
            data={"name": "ApplicationUpdate"},
            method=HTTPMethod.POST,
        )

    async def async_command_backup(self) -> HTTPResponse:
        """Trigger a backup routine."""
        return await self._async_request(
            "command",
            data={"name": "Backup"},
            method=HTTPMethod.POST,
        )

    async def async_command_check_health(self) -> HTTPResponse:
        """Trigger a system health check."""
        return await self._async_request(
            "command",
            data={"name": "CheckHealth"},
            method=HTTPMethod.POST,
        )

    async def async_command_clear_blocklist(self) -> HTTPResponse:
        """Trigger the removal of all blocklisted movies."""
        return await self._async_request(
            "command",
            data={"name": "ClearBlocklist"},
            method=HTTPMethod.POST,
        )

    async def async_command_clean_recycle_bin(self) -> HTTPResponse:
        """Trigger a recycle bin cleanup check."""
        return await self._async_request(
            "command",
            data={"name": "CleanUpRecycleBin"},
            method=HTTPMethod.POST,
        )

    async def async_command_delete_log_files(self) -> HTTPResponse:
        """Trigger the removal of all Info/Debug/Trace log files."""
        return await self._async_request(
            "command",
            data={"name": "DeleteLogFiles"},
            method=HTTPMethod.POST,
        )

    async def async_command_delete_update_log_files(self) -> HTTPResponse:
        """Trigger the removal of all Update log files."""
        return await self._async_request(
            "command",
            data={"name": "DeleteUpdateLogFiles"},
            method=HTTPMethod.POST,
        )

    async def async_command_downloaded_movies_scan(self) -> HTTPResponse:
        """Trigger the scan of downloaded movies."""
        return await self._async_request(
            "command",
            data={"name": "DownloadedMoviesScan"},
            method=HTTPMethod.POST,
        )

    async def async_command_missing_movies_search(self) -> HTTPResponse:
        """Trigger a search of all missing movies."""
        return await self._async_request(
            "command",
            data={"name": "MissingMoviesSearch"},
            method=HTTPMethod.POST,
        )

    async def async_command_refresh_monitored_downloads(self) -> HTTPResponse:
        """Trigger the scan of monitored downloads."""
        return await self._async_request(
            "command",
            data={"name": "RefreshMonitoredDownloads"},
            method=HTTPMethod.POST,
        )

    async def async_command_refresh_movies(self, movieids: list[int]) -> HTTPResponse:
        """Trigger a refresh / scan of movies."""
        return await self._async_request(
            "command",
            data={"name": "RefreshMovie", "movieIds": movieids},
            method=HTTPMethod.POST,
        )

    async def async_get_calendar(
        self, start_date: datetime, end_date: datetime, unmonitored: bool = True
    ) -> list[RadarrCalendar]:
        """Get a list of movies based on calendar parameters."""
        params = {
            "start": start_date.strftime("%Y-%m-%d"),
            "end": end_date.strftime("%Y-%m-%d"),
            "unmonitored": str(unmonitored),
        }
        return await self._async_request(
            "calendar",
            params=params,
            datatype=RadarrCalendar,
        )

    @api_command("system/status", datatype=RadarrSystemStatus)
    async def async_get_system_status(self) -> RadarrSystemStatus:
        """Get information about system status."""

    @api_command("health", datatype=RadarrHealth)
    async def async_get_failed_health_checks(self) -> RadarrHealth:
        """Get information about failed health checks."""

    @api_command("update", datatype=RadarrUpdate)
    async def async_get_software_update_info(self) -> RadarrUpdate:
        """Get information about software updates."""

    @api_command("qualityProfile", datatype=RadarrQualityProfile)
    async def async_get_quality_profiles(self) -> list[RadarrQualityProfile]:
        """Get information about quality profiles."""

    @api_command("customfilter", datatype=RadarrCustomFilter)
    async def async_get_custom_filters(self) -> list[RadarrCustomFilter]:
        """Get information about custom filters."""

    @api_command("remotePathMapping", datatype=RadarrRemotePathMapping)
    async def async_get_remote_path_mappings(self) -> list[RadarrRemotePathMapping]:
        """Get information about remote path mappings."""

    @api_command("rootfolder", datatype=RadarrRootFolder)
    async def async_get_root_folders(self) -> list[RadarrRootFolder]:
        """Get information about root folders."""

    async def _async_construct_movie_json(  # pylint: disable=too-many-arguments
        self,
        db_id: str,
        quality_profile_id: int,
        root_dir: str,
        monitored: bool = True,
        search_for_movie: bool = True,
        tmdb: bool = True,
    ) -> dict[str, str | int | dict[str, bool]]:
        """Search for movie on tmdb and returns Movie json to add.

        Args:
            db_id: imdb or tmdb id
            quality_profile_id: ID of the quality profile the movie will use
            root_dir: location of the root DIR
            monitored: should the movie be monitored.
            search_for_movie: Should we search for the movie.
            tmdb: Use TMDB IDs. Set to False to use IMDB.
        """
        if not (movie := await self.async_lookup_movie(db_id, tmdb=tmdb)[0]):
            raise Exception("Movie Doesn't Exist")

        return {
            "title": movie["title"],
            "rootFolderPath": root_dir,
            "qualityProfileId": quality_profile_id,
            "year": movie["year"],
            "tmdbId": movie["tmdbId"],
            "images": movie["images"],
            "titleSlug": movie["titleSlug"],
            "monitored": monitored,
            "addOptions": {"searchForMovie": search_for_movie},
        }



"""API response model for AioPyArr Api."""
from __future__ import annotations

from enum import Enum
from typing import Any

from .base import APIResponseType, BaseModel


class APIResult(str, Enum):
    """ApiResult."""

    SUCCESS = "success"
    ERROR = "error"


class PyArrResponse(BaseModel):  # pylint: disable=too-few-public-methods
    """API response model for PyArr Api."""

    data: dict[str, Any] | list[dict[str, Any]] | None = None
    message: str | None = None
    result: APIResult | None = None

    def _generate_data(
        self, data: dict[str, Any] | list[dict[str, Any]]
    ) -> dict[str, Any] | list[dict[str, Any]]:
        """Generate data."""
        if self._datatype is None:
            return data

        if (
            self._datatype._responsetype  # pylint: disable=protected-access
            == APIResponseType.LIST
        ):
            return [self._datatype(item, self._datatype) for item in data]

        return self._datatype(data, self._datatype)



"""PyArr model constants."""

CONVERT_TO_BOOL = (
    "Advanced",
    "analyticsEnabled",
    "approved",
    "authentication",
    "downloadAllowed",
    "downloading",
    "enable",
    "enableAuto",
    "enableAutomaticSearch",
    "enableColorImpairedMode",
    "enabled",
    "enableInteractiveSearch",
    "enableRss",
    "enableSsl",
    "errors",
    "fullSeason",
    "hasFile",
    "includeCustomFormatWhenRenaming",
    "includeHealthWarnings",
    "includeQuality",
    "installable",
    "installed",
    "isAbsoluteNumbering",
    "isAdmin",
    "isAvailable",
    "isDaily",
    "isDebug",
    "isDocker",
    "isLinux",
    "isMono",
    "isNetCore",
    "isNewMovie",
    "isOsx",
    "isPossibleSpecialEpisode",
    "isProduction",
    "isRepack",
    "isUserInteractive",
    "isWindows",
    "launchBrowser",
    "latest",
    "monitored",
    "negate",
    "onDelete",
    "onDownload",
    "onGrab",
    "onHealthIssue",
    "onRename",
    "onUpgrade",
    "proper",
    "proxyBypassLocalAddresses",
    "proxyEnabled",
    "qualityCutoffNotMet",
    "renameMovies",
    "replaceIllegalCharacters",
    "replaceSpaces",
    "required",
    "sceneSource",
    "searchOnAdd",
    "sendUpdatesToClient",
    "shouldMonitor",
    "showRelativeDates",
    "special",
    "supportsOnDelete",
    "supportsOnDownload",
    "supportsOnGrab",
    "supportsOnHealthIssue",
    "supportsOnRename",
    "supportsOnUpgrade",
    "supportsRss",
    "supportsSearch",
    "unknownErrors",
    "unknownWarnings",
    "unverifiedSceneNumbering",
    "updateAutomatically",
    "warnings",
)



"""PyArr constants."""
from enum import Enum
from logging import Logger, getLogger

LOGGER: Logger = getLogger(__package__)

ATTR_DATA = "data"


class HTTPMethod(Enum):
    """HTTPMethod Enum."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


class HTTPResponse(Enum):
    """HTTPResponse Enum."""

    OK = 200
    UNAUTHORIZED = 401
    RESOURCENOTFOUND = 404



"""Decorator for pyarr."""
from __future__ import annotations

from typing import TYPE_CHECKING

from .const import HTTPMethod
from .models.base import BaseModel

if TYPE_CHECKING:
    from .request_client import RequestClient


def api_command(
    command: str,
    params: dict | None = None,
    data: dict | None = None,
    datatype: BaseModel | None = None,
    method: HTTPMethod = HTTPMethod.GET,
):
    """Initialize decorator for PyArr API request."""

    def decorator(_):
        """Initialize decorator."""

        async def wrapper(*args):
            """Initialize wrapper."""
            client: RequestClient = args[0]
            return await client._async_request(  # pylint: disable=protected-access
                *args,
                command,
                params=params,
                data=data,
                datatype=datatype,
                method=method
            )

        return wrapper

    return decorator
